;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; $Id: enigma.m4,v 1.5 2005/11/08 04:15:02 markv Exp $
;;;
;;; enigma2600.m4
;;;
;;; a replica of the German three rotor enigma machine, coded for the
;;; Atari 2600, a video game of the late 70's and early 80's.  Every
;;; effort has been made to make this authentic: it should be able to
;;; decode messages from real Enigma machines.
;;;
;;; Copyright 2005, Mark VandeWettering <mvandewettering@gmail.com>
;;; 
;;; Feel free to download, execute, modify, share and even sell this 
;;; code as long as you retain the copyright above.  
;;;
;;; This is the first real program I've ever written for the Atari 2600,
;;; and the first code that I've written for the 6502 in probably twenty
;;; years.
;;;
;;; It's been coded to use the perl version of p65, a perl assembler.  
;;; Because the version I had did not include macro capabilities, I've
;;; augmented it with macros which were written using the standard Unix
;;; macro processor m4.
;;;
;;; This code actually uses so called "F8 bankswitching" for two reasons:
;;; firstly, I was lazy and didn't feel like trying to wedge this code
;;; down to 4K of space, and secondly because it will serve as documentation
;;; for how you can use bank switched cartridge.   Well, that, and the 
;;; EEPROM chips that could find were 8Kx8 chips (AT28C64B).
;;;
;;; $Log: enigma.m4,v $
;;; Revision 1.5  2005/11/08 04:15:02  markv
;;; Version that I'm committing to use at Hackers 2005
;;;
;;; Revision 1.4  2005/10/31 02:26:41  markv
;;; I believe the code is approaching 100% functional.  All of the interface
;;; elements are there and work, only the stecker verification routine remains
;;; to be written.
;;;
;;; Revision 1.3  2005/10/31 02:08:32  markv
;;; I managed to lose this code once, and had to reconstruct it all over again
;;; from scratch.  I'm saving it just to be safe.  It mostly works, some minor
;;; display and interface issues could be better, and we need to verify the
;;; plugboard settings.   Then, maybe an attract mode?   Or perhaps a colorful
;;; banner page?
;;;
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	.include "vcs.h"

;;;
;;; data 
;;;
	.data
	.org $80
	.space grheight 1
	.space tmpvar	1
	.space LETBASE	12		; points to six letters
	.alias LET_1	LETBASE
	.alias LET_2	LETBASE+2
	.alias LET_3	LETBASE+4
	.alias LET_4	LETBASE+6
	.alias LET_5	LETBASE+8
	.alias LET_6	LETBASE+10
	.space COL	1		; cursor position
	.space ROW	1
	.space tmprow	1
	.space counter	1
	.space IP	2

	.space STATE		1
	.space NEXTSTATE	1
	.space DELAY		1
	
	.space FWD	6	; three pointers to the forward rings...
	.alias FWD_1	FWD+0
	.alias FWD_2	FWD+2
	.alias FWD_3	FWD+4

	.space REV	6	; and the pointers to the reverse rings...
	.alias REV_1	REV+0
	.alias REV_2	REV+2
	.alias REV_3	REV+4

	.space KEY	3	; the current key settings
	.alias KEY_1	KEY+0
	.alias KEY_2	KEY+1
	.alias KEY_3	KEY+2

	.space RING	3	; the current ring settings
	.alias RING_1	RING+0
	.alias RING_2	RING+1
	.alias RING_3	RING+2

	.space NOTCH	3	; the ring notch locations
	.alias NOTCH_1	NOTCH+0
	.alias NOTCH_2	NOTCH+1
	.alias NOTCH_3	NOTCH+2

	.space ROTOR	3	; which rotor is loaded in each position
	.alias ROTOR_1	ROTOR+0
	.alias ROTOR_2	ROTOR+1
	.alias ROTOR_3	ROTOR+2
	
	;;; stecker settings... 
	;;; we allow up to six plugs, with the idea being that 
	;;; if a character matches FWD_STK[n], then it is replaced by
	;;; REV_STK[n] _and vice versa_.

	.space STK	12
	.alias FWD_STK	STK
	.alias REV_STK	STK+6

	.space INPUT	1	; current input character
	.space OUTPUT	1	; the result of encoding input
	.space INBUF	6	; buffer for the input chars
	.space OUTBUF	6	; buffer for the output chars

define(`DELAY',6)

define(`init', `sei
	cld
	ldx #$ff
	txs
	lda #0
*	sta 0,x
	dex
	bne -')dnl

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define(`verticalsync',`lda #2
	sta vsync
	ldx #3
	jsr doblank
	lda #0
	sta vsync')dnl

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define(`startvblank',`; set timers to begin vertical blank;
	lda `#'eval((76*38-12)/64)
	sta tim64t')dnl

define(`finishvblank',`; wait for vertical blank timer to expire
	; then do a store to wsync to synchronize 
*	lda intim
	bne -
	lda #0
	sta wsync
	sta vblank
')dnl

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define(`setpos',`; set position
	sta hmclr
	sec
	sta wsync
*	sbc #15
	bcs -
	eor #7
	asl
	asl
	asl
	asl
	;; the following is actually a sta hmp0,x but with an 
	;; absolute argument, rather than a zero page.  It must
	;; take five cycles to work, and p65 wont generate the
	;; instruction directly.
	.byte 157
	.byte hmp0
	.byte 0
	sta resp0,x
	sta wsync
	sta hmove')dnl
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define(`keyline',`; setup the ram tables...
	sta wsync
	lda #0
	sta enabl
	lda tmprow
	cmp `#'$7
	bne +
	lda #2
	sta enabl
*	lda #<letter_$1
	sta LET_1
	lda #<letter_$2
	sta LET_2
	lda #<letter_$3
	sta LET_3
	lda #<letter_$4
	sta LET_4
	lda #<letter_$5
	sta LET_5
	lda #<letter_$6
	sta LET_6

	lda #5
	sta grheight
	jsr drawsix')dnl

define(`bufline', `; setup the tables...
	ldy $1
	lda letter_lo,y
	sta LET_1
	ldy $1+1
	lda letter_lo,y
	sta LET_2
	ldy $1+2
	lda letter_lo,y
	sta LET_3
	ldy $1+3
	lda letter_lo,y
	sta LET_4
	ldy $1+4
	lda letter_lo,y
	sta LET_5
	ldy $1+5
	lda letter_lo,y
	sta LET_6
	lda #5
	sta grheight
	jsr drawsix')dnl

define(`stkline', `; setup the tables...
	sta wsync
	lda #0
	sta enabl
	lda tmprow
	cmp `#'$2
	bne +
	lda #2
	sta enabl
*
	ldy $1
	lda letter_lo,y
	sta LET_1
	ldy $1+1
	lda letter_lo,y
	sta LET_2
	ldy $1+2
	lda letter_lo,y
	sta LET_3
	ldy $1+3
	lda letter_lo,y
	sta LET_4
	ldy $1+4
	lda letter_lo,y
	sta LET_5
	ldy $1+5
	lda letter_lo,y
	sta LET_6
	lda #5
	sta grheight
	jsr drawsix')dnl

define(`posline',`; setup the ram tables...

	sta wsync
	lda #0
	sta enabl
	lda state
	cmp #$4
	bne +
	lda #2
	sta enabl
* 	ldy $1
	lda letter_lo,y
	sta LET_1
	ldy $2
	lda letter_lo,y
	sta LET_2
	ldy $3
	lda letter_lo, y
	sta LET_3

	lda #5
	sta grheight
	jsr drawthree
	sta wsync	
	lda #0
	sta grp0
	sta grp1
	sta grp0
	sta grp1')dnl

define(`rtrline',`; setup the ram tables...
	sta wsync
	lda #0
	sta enabl
	lda state
	cmp #STATE_ROTOR
	bne +
	lda #2
	sta enabl
* 	ldy rotor_3
	lda digit_lo,y
	sta LET_1
	lda digit_hi,y
	sta LET_1+1
	ldy rotor_2
	lda digit_lo,y
	sta LET_2
	lda digit_hi,y
	sta LET_2+1
	ldy rotor_1
	lda digit_lo, y
	sta LET_3
	lda digit_hi,y
	sta LET_3+1

	lda #5
	sta grheight
	jsr drawthree

	lda #64
	sta hmp0
	sta wsync
	sta hmove
	lda #0
	sta grp0
	sta grp1
	sta grp0
	sta grp1
	sta wsync
	sta enabl')dnl
	

define(`steckerblock',`;
	jsr poscursor
	lda ROW
	sta tmprow
	lda state
	cmp #STATE_PLUG
	beq +
	lda #2
	sta tmprow
*	
	stkline(FWD_STK, 0)
	stkline(REV_STK, 1)
	lda #0
	sta wsync
	sta enabl')dnl


define(`letterblock',`;
	jsr poscursor
	lda ROW
	sta tmprow
	lda state
	cmp #STATE_ENCODE
	beq +
	lda #10
	sta tmprow
*
	keyline(a,b,c,d,e,f,0)
	keyline(g,h,i,j,k,l,1)
	keyline(m,n,o,p,q,r,2)
	keyline(s,t,u,v,w,x,3)
	keyline(bl,bl,y,z,bl,bl,4)
	lda #0
	sta wsync
	sta enabl')dnl

define(`drawbanner',`
	lda `#$'A4
	sta wsync
	sta hmclr
	sta colubk
	lda `#$'4e
	sta colup0
	sta colup1
	lda #1
	sta vdelp0
	sta vdelp1
	lda #0
	sta grp0
	sta grp1
	sta grp0
	sta grp1
	lda #5
	sta grheight
* 	ldy grheight	; (+3)
	lda $1_1,y	; (+7)
	sta grp0	; (+10)
	sta wsync	; (+13)

	;; P0 at 124,140,156	(cycle 41.33, 46.66, 52,00)
	;; P1 at 132,148,164	(cycle 44.00, 49.33, 54.67)

	lda $1_2,y	; (+4)
	sta grp1	; (+7)		P0 -> letter_a visible
	lda $1_3,y	; (+11)
	sta grp0	; (+14)		P1 -> letter_b visible
	lda $1_4,y  ; (+18)		
	sta tmpvar	; (+21)
	lda $1_5,y	; (+25)
	tax		; (+27)
	lda $1_6,y	; (+31)
	tay		; (+33)
	lda tmpvar	; (+36)
	nop		; (+38)
	bit 0		; (+41)
	sta grp1	; (+44)
	stx grp0	; (+45)
	sty grp1	; (+48)
	sta grp0	; (+51)
	dec grheight	; (+56)
	bpl -		; (+59)

	sta wsync
	lda #0
	sta grp0
	sta grp1
	sta grp0
	sta grp1
	lda `#$'88
	sta wsync
	sta colubk
	lda `#$'0E
	sta colup0
	sta colup1')dnl

define(`drawdoublebanner',`
	lda counter
	sta wsync
	sta colubk
	lda `#$'4e
	sta colup0
	sta colup1
	lda #1
	sta vdelp0
	sta vdelp1
	lda #0
	sta grp0
	sta grp1
	sta grp0
	sta grp1
	lda #12
	sta grheight
* 	ldy grheight	; (+3)
	lda $1_1,y	; (+7)
	sta grp0	; (+10)
	sta wsync	; (+13)

	;; P0 at 124,140,156	(cycle 41.33, 46.66, 52,00)
	;; P1 at 132,148,164	(cycle 44.00, 49.33, 54.67)

	lda $1_2,y	; (+4)
	sta grp1	; (+7)		P0 -> letter_a visible
	lda $1_3,y	; (+11)
	sta grp0	; (+14)		P1 -> letter_b visible
	lda $1_4,y  ; (+18)		
	sta tmpvar	; (+21)
	lda $1_5,y	; (+25)
	tax		; (+27)
	lda $1_6,y	; (+31)
	tay		; (+33)
	lda tmpvar	; (+36)
	nop		; (+38)
	bit 0		; (+41)
	sta grp1	; (+44)
	stx grp0	; (+45)
	sty grp1	; (+48)
	sta grp0	; (+51)
	dec grheight	; (+56)
	bpl -		; (+59)

	sta wsync
	lda #0
	sta grp0
	sta grp1
	sta grp0
	sta grp1

	lda `#$'88
	sta wsync
	sta colubk
	lda `#$'0E
	sta colup0
	sta colup1')dnl

	
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	.org $F000
page0_dofsm:
	sta $FFF9
	nop		; blank space, filled by jsr...
	nop
	nop
	nop		; blank space, filled by the sta
	nop
	nop
	rts		; we are back from page zero.

main:
	sta $FFF8		; make sure we are executing from bank zero
	init

	lda #$0
	sta colupf
	lda #$88
	sta colubk

	lda #$0E
	sta colup0
	sta colup1
	
	lda #$ff
	sta grp0
	sta grp1

	lda #3
	sta nusiz0
	sta nusiz1

	lda #$ff	; put some border in to hide the hmoves...
	sta pf0
	sta pf1
	lda #%00000001
	sta pf2

	lda #$31	; reflect the playfield
	sta ctrlpf	

	lda #0
	sta hmclr

	lda #>letter_a
	sta LET_1+1
	sta LET_2+1
	sta LET_3+1
	sta LET_4+1
	sta LET_5+1
	sta LET_6+1

	lda #0
	sta input
	sta row
	sta col

	lda #STATE_ROTOR
	sta STATE

define(`COUNTERSPEED', 2)

	lda `#'COUNTERSPEED
	sta counter

	lda #2
	sta ROTOR_1
	lda #1
	sta ROTOR_2
	lda #0
	sta ROTOR_3

define(`store3',`;;; store three...
	lda #$2
	sta $1_1
	lda #$3
	sta $1_2
	lda #$4
	sta $1_3')
	
	store3(key,0,0,0)
	store3(ring,0,0,0)

	ldy #5
	lda #26
*	sta OUTBUF,y
	sta INBUF,y
	dey
	bpl -

	ldy #5
*	tya
	sta FWD_STK,y
	sta REV_STK,y
	dey
	bpl -
	
kernel:
	verticalsync

	startvblank

	jsr page1_dofsm

	ldx #0
	lda #56
	setpos			; 2 lines
	ldx #1
	lda #64
	setpos			; 2 lines


	finishvblank

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; 192 lines of visible screen...
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

				; 8 lines
	lda COL
	asl
	asl
	asl
	asl
	eor #$ff
	clc 
	adc #94
	ldx #4
	jsr blackpos

	drawdoublebanner(enigma) 

	ldx #7
	jsr doblank

	drawbanner(rotors)
	lda #$C0
	sta hmp0
	sta wsync
	sta hmove
	rtrline()			; 2 lines

	drawbanner(rings)
	lda #$C0
	sta hmp0
	sta wsync
	sta hmove
	posline(RING_3, RING_2, RING_1, STATE_RING)	; 2 lines
	lda #$40
	sta hmp0
	sta wsync
	sta hmove
	lda #0
	sta enabl
	drawbanner(keybanner)
	lda #$C0
	sta hmp0
	sta wsync
	sta hmove
	posline(KEY_3, KEY_2, KEY_1, STATE_KEY)	; 2 lines
	lda #$40
	sta hmp0
	sta wsync
	sta hmove
	lda #0
	sta enabl

	drawbanner(plugs)
	steckerblock

	drawbanner(input)
	bufline(INBUF)
	letterblock			; 40 lines

	drawbanner(output)
	bufline(OUTBUF)
	
	ldx #2
	jsr doblank
	
	lda #2
	sta vblank

	ldx #30
	jsr doblank

	jmp kernel

doblank:
*	sta wsync
	dex
	bne -
	rts
	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

blackblank:
	lda #0
	sta wsync
	sta colubk	
	lda #$88
	sta wsync
	sta colubk
	rts

doubleblack:
	lda #0
	sta wsync
	sta colubk
	lda #$88	
	sta wsync
	sta wsync
	sta colubk
	rts
	
blackpos:
	sta hmclr
	sec
	sta wsync
*	sbc #15
	bcs -
	eor #7
	asl
	asl
	asl
	asl
	;; the following is actually a sta hmp0,x but with an 
	;; absolute argument, rather than a zero page.  It must
	;; take five cycles to work, and p65 wont generate the
	;; instruction directly.
	.byte 157
	.byte hmp0
	.byte 0
	sta resp0,x
	sta wsync
	sta hmove
	rts

drawsix:
	ldy grheight
	lda (LET_1),y
	sta grp0
	sta wsync
	lda (LET_2),y
	sta grp1
	lda (LET_3),y
	sta grp0
	lda (LET_4),y
	sta tmpvar
	lda (LET_5),y
	tax
	lda (LET_6),y
	tay
	lda tmpvar
	sta grp1
	stx grp0
	sty grp1
	sta grp0
	dec grheight
	bpl drawsix
	sta wsync	
	lda #0
	sta grp0
	sta grp1
	sta grp0
	sta grp1
	rts

;;;
;;; p0 appears at pixel 60-68, 76-84, and 92-100
;;; corresponds to cpu cycle 42.6-45.3, 48-50.6 and 53.3-122.6

drawthree:
	sta wsync
	ldy grheight	; +3 +3
	ldx #0		; +2 +5
	lda (LET_1),y	; +5 +10
	sta grp0	; +3 +13
	stx grp1	; +3 +16   LET1 VISIBLE
	lda (LET_2),y	; +5 +21 
	sta grp0	; +3 +24
	lda (LET_3),y	; +5 +29
	bit 0		; +3 +32
	bit 0		; +3 +35
	bit 0		; +3 +38
	bit 0		; +3 +41
	nop		; +2 +43
	stx grp1	; +3 +46
	sta grp0	; +3 +48
	stx grp1	; +3 +51
	stx grp0	; +3 +54
	dec grheight	; +5 +59
	bpl drawthree	; +3 +62
	rts

	.include "new.h"

poscursor:
	lda COL
	asl
	asl
	asl
	clc
	adc #57
	ldx #4
	jmp blackpos
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


	.advance $FE00
	.include "atariletters.h"
	.include "digits.h"

	.advance $FFFC
	.word main
	.word main

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; I've decided to move all the having to do with handling the state 
;;; machine (the code which normally gets called in the vertical blank
;;; period) to be moved to the second page.  This should leave plenty of
;;; room in the first page for all the display related items.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	.org $F000
page1_dofsm:
	sta $FFF9		; we are executing from bank one...
	jsr do_fsm		; execute the state machine
	sta $FFF8		; back to page zero...
	rts
shadowmain:
	;;; make sure we execute code in page zero, regardless of the
	;;; status of the flip-flop.
	sta $FFF8		; execute from bank zero.

;;;
;;;
;;; do_fsm is the main entry point that drives the finite state machine,
;;; it is called in the vertical blank portion to actually handle all
;;; the reading of the joystick, switches and button and drive the actual
;;; computations.   
;;;
do_fsm:
	ldx delay		; if delay is not zero, then decrement
	bne do_fsm_delay	; and skip further state machine stuff.
	ldx state		; otherwise, use the state to index the
	lda statelo, x		; jmp table and do an indirect jmp
	sta ip			; this effectively makes it look like
	lda statehi,x		; we are calling the various state functions
	sta ip+1		; as subroutines.
	jmp (ip)

do_fsm_delay:
	dex
	stx delay
	rts

.alias STATE_ATTRACT 		0
.alias STATE_ROTOR		1
.alias STATE_RING		2
.alias STATE_PLUG		3
.alias STATE_KEY		4
.alias STATE_ENCODE		5
.alias STATE_PERFORM_ENCODE	6
.alias STATE_VERIFY_ROTOR	7
.alias STATE_VERIFY_PLUG	8
.alias STATE_ERROR		9

.alias JOY_RIGHT	%10000000
.alias JOY_LEFT		%01000000
.alias JOY_DOWN		%00100000
.alias JOY_UP		%00010000

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_attract:
	lda #STATE_ROTOR
	sta state
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_rotor:
	lda #0			; we could be back here because of error
	sta audv0		; so turn off the buzzer...

	lda #JOY_RIGHT
	bit swcha
	bne +
	lda COL
	clc
	adc #2
	tay
	lda mod3,y
	sta COL
*
	lda #JOY_LEFT
	bit swcha
	bne +
	lda COL
	clc
	adc #1
	tay	
	lda mod3,y	
	sta COL
*
	lda #JOY_UP
	bit swcha
	bne +
	ldx COL
	lda ROTOR,x
	clc
	adc #1
	tay
	lda mod5,y
	sta ROTOR,x
*
	lda #JOY_DOWN
	bit swcha
	bne +
	ldx COL
	lda ROTOR,x
	clc
	adc #4
	tay
	lda mod5,y
	sta ROTOR,x
*
	lda #2
	bit swchb		; if the select is pressed..
	bne +			; try verifying the rotors
	lda #STATE_VERIFY_ROTOR	; next time around...
	sta state
	lda #0
	sta input
	sta row
	sta col
*
	lda #10
	sta delay
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_ring:
	lda #JOY_RIGHT
	bit swcha
	bne +
	lda COL
	clc
	adc #2
	tay
	lda mod3,y
	sta COL
*
	lda #JOY_LEFT
	bit swcha
	bne +
	lda COL
	clc
	adc #1
	tay	
	lda mod3,y	
	sta COL
*
	lda #JOY_UP
	bit swcha
	bne +
	ldx COL
	lda RING,x
	clc
	adc #1
	tay
	lda mod26,y
	sta RING,x
*
	lda #JOY_DOWN
	bit swcha
	bne +
	ldx COL
	lda RING,x
	clc
	adc #25
	tay
	lda mod26,y
	sta RING,x
*
	lda #2
	bit swchb
	bne +
	lda #STATE_KEY
	sta state
	lda #0
	sta input
	sta row
	sta col
*
	lda #10
	sta delay
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_key:
	lda #JOY_RIGHT
	bit swcha
	bne +
	lda COL
	clc
	adc #2
	tay
	lda mod3,y
	sta COL
*
	lda #JOY_LEFT
	bit swcha
	bne +
	lda COL
	clc
	adc #1
	tay	
	lda mod3,y	
	sta COL
*
	lda #JOY_UP
	bit swcha
	bne +
	ldx COL
	lda KEY,x
	clc
	adc #1
	tay
	lda mod26,y
	sta KEY,x
*
	lda #JOY_DOWN
	bit swcha
	bne +
	ldx COL
	lda KEY,x
	clc
	adc #25
	tay
	lda mod26,y
	sta KEY,x
*
	lda #2
	bit swchb
	bne +
	lda #STATE_PLUG
	sta state
	lda #0
	sta input
	sta row
	sta col
*
	lda #10
	sta delay
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_plug:
	lda #0				; silence the bleeps 
	sta audv0

	lda #JOY_RIGHT
	bit swcha
	bne +
	lda input
	clc
	adc #1
	tay
	lda mod12,y
	sta input
*
	lda #JOY_LEFT
	bit swcha
	bne +
	lda input
	clc
	adc #11
	tay
	lda mod12,y
	sta input
*
	lda #JOY_UP
	bit swcha
	bne +
	ldx input
	lda STK,x
	clc
	adc #1
	tay
	lda mod26,y
	sta STK,x
*
	lda #JOY_DOWN
	bit swcha
	bne +
	ldx input
	lda STK,x
	clc
	adc #25
	tay
	lda mod26,y
	sta STK,x
*
	lda #2
	bit swchb
	bne +
	lda #STATE_VERIFY_PLUG
	sta state
*
	jmp setrowtabs

setrowtabs:
	ldy input
	lda rowtab,y
	sta ROW
	lda coltab,y
	sta COL
	;;; falls through!  be careful!	
delayrts:
	lda #10
	sta delay
	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


do_encode:
	lda #0				; silence the bleeps 
	sta audv0

	lda #JOY_LEFT
	bit swcha
	bne +
	lda input
	clc
	adc #25
	tay
	lda mod26,y
	sta input
*	
	lda #JOY_RIGHT
	bit swcha
	bne +
	lda input
	clc
	adc #1
	tay
	lda mod26,y
	sta input
*	
	lda #2				; select moves you back to rings...
	bit swchb
	bne +
	lda #STATE_ROTOR
	sta state
	lda #0
	sta input
	sta row
	sta col
*
	lda inpt4
	bmi +
	lda #STATE_PERFORM_ENCODE	; jmp to perform the encode
	sta state
	jsr inputchar			; add the char to the buffer
*
	jmp setrowtabs 


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_perform_encode:
	jsr advance
	lda input
	jsr encode			; encode a single character
	sta output
	sta audf0			; make a frequency...
	jsr outputchar
	lda #STATE_ENCODE		; and then jmp back to moving
	sta state

	;;; make some bleeps and bloops
	lda #%00000110
	sta audc0
	lda #$8
	sta audv0
	
	jmp delayrts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_error:
	lda nextstate
	sta state
	rts

do_verify_rotor:
	lda ROTOR
	cmp ROTOR+1
	beq bad_rotor
	cmp ROTOR+2
	beq bad_rotor
	lda ROTOR+1
	cmp ROTOR+2
	beq bad_rotor
	;;; all good, we can go on to STATE_RING

	jsr page1loadrotors

	lda #STATE_RING
	sta state
	rts
bad_rotor:
	lda #STATE_ROTOR
	sta nextstate
	lda #STATE_ERROR
	sta state
	lda #%00000111
	sta audf0
	lda #%00000110
	sta audc0
	lda #$8
	sta audv0

	jmp delayrts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

do_verify_plug:
	;;;
	;;; check to see if the plugboard is setting is valid
	;;; A setting is valid if the each letter in FWD_STK and REV_STK
	;;; are unique, or if FWD_STK[i] == REV_STK[i] 
	;;;
	ldy #11
plugloop:	
	tya
	pha		
	ldx #1
	lda STK,y
	dey
	bmi trynext
*	cmp STK,y
	bne +
	inx
*	dey
	bpl --

	cpx #1
	beq trynext

	cpx #2
	bne bad_plug
	
	pla
	pha
	tay
	lda mod6,y
	tay
	lda FWD_STK,y
	cmp REV_STK,y
	bne bad_plug
trynext:
	pla
	tay
	dey
	bpl plugloop

	; everything seemed fine...

	lda #0
	sta input
	sta row
	sta col
	lda #STATE_ENCODE
	sta state
	rts

bad_plug:
	pla 		; we still have a count on the stack...
	lda #STATE_PLUG
	sta nextstate
	lda #STATE_ERROR
	sta state
	lda #%00000111
	sta audf0
	lda #%00000110
	sta audc0
	lda #$8
	sta audv0
	jmp delayrts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define(`rotate',`clc
	adc #26		
	sec
	sbc $2		; subtract the ring position
	tay
	lda mod26,y
	clc
	adc $3		; add in the position
	tay
	lda mod26,y
	tay		; now index to
	lda ($1),y	; go through the rotor
	clc
	adc $2		; add back in the ring 
	tay
	lda mod26,y	; and mod
	clc
	adc #26
	sec
	sbc $3		; subtract out the pos
	tay
	lda mod26,y')

;;;
;;; Here are the routines which actually do the simulation of the 
;;; Enigma.  Note that the rotor advance occurs prior to the coding,
;;; and that the rotors don't step in a strictly odometer like fashion.

advance:
	lda KEY_2
	cmp NOTCH_2
	;;; advance both rotors...
	bne dontadvance3
	ldy KEY_2
	iny
	lda mod26,y
	sta KEY_2
	ldy KEY_3
	iny
	lda mod26,y
	sta KEY_3
	jmp dontadvance2
dontadvance3:
	lda KEY_1
	cmp NOTCH_1
	bne dontadvance2
	ldy KEY_2
	iny
	lda mod26,y
	sta KEY_2
dontadvance2:
	ldy KEY_1	;;; always increment rotor 
	iny
	lda mod26,y
	sta KEY_1
	rts

encode:
	jsr dostecker
	rotate(FWD_1,RING_1,KEY_1)
	rotate(FWD_2,RING_2,KEY_2)
	rotate(FWD_3,RING_3,KEY_3)
	tay
	lda REFLECT1,y
	rotate(REV_3,RING_3,KEY_3)
	rotate(REV_2,RING_2,KEY_2)
	rotate(REV_1,RING_1,KEY_1)
	jsr dostecker
	rts

dostecker:
	ldy #5
*	cmp FWD_STK,y
	beq found_fwd
	dey
	bpl -
	ldy #5
*	cmp REV_STK,y
	beq found_rev
	dey
	bpl -
	rts
found_fwd:
	lda REV_STK,y
	rts
found_rev:
	lda FWD_STK,y
	rts

inputchar:
	ldy #1
*	lda INBUF,y
	sta INBUF-1,y
	iny
	cpy #6
	bcc -
	lda input
	sta INBUF+5
	rts

outputchar:
	ldy #1
*	lda OUTBUF,y
	sta OUTBUF-1,y
	iny
	cpy #6
	bcc -
	lda output
	sta OUTBUF+5
	rts

page1loadrotors:
	ldx ROTOR_1
	lda FROT_lo,x
	sta FWD_1	
	lda FROT_hi,x
	sta FWD_1+1
	lda RROT_lo,x
	sta REV_1	
	lda RROT_hi,x
	sta REV_1+1
	lda NOTCH_TAB,x
	sta NOTCH_1

	ldx ROTOR_2
	lda FROT_lo,x
	sta FWD_2	
	lda FROT_hi,x
	sta FWD_2+1
	lda RROT_lo,x
	sta REV_2	
	lda RROT_hi,x
	sta REV_2+1
	lda NOTCH_TAB,x
	sta NOTCH_2

	ldx ROTOR_3
	lda FROT_lo,x
	sta FWD_3
	lda FROT_hi,x
	sta FWD_3+1
	lda RROT_lo,x
	sta REV_3	
	lda RROT_hi,x
	sta REV_3+1
	lda NOTCH_TAB,x
	sta NOTCH_3

	rts


	ldx ROTOR_1
	lda FROT_lo,x
	sta FWD_1	
	lda FROT_hi,x
	sta FWD_1+1
	lda RROT_lo,x
	sta REV_1	
	lda RROT_hi,x
	sta REV_1+1
	lda NOTCH_TAB,x
	sta NOTCH_1

	ldx ROTOR_2
	lda FROT_lo,x
	sta FWD_2	
	lda FROT_hi,x
	sta FWD_2+1
	lda RROT_lo,x
	sta REV_2	
	lda RROT_hi,x
	sta REV_2+1
	lda NOTCH_TAB,x
	sta NOTCH_2

	ldx ROTOR_3
	lda FROT_lo,x
	sta FWD_3
	lda FROT_hi,x
	sta FWD_3+1
	lda RROT_lo,x
	sta REV_3	
	lda RROT_hi,x
	sta REV_3+1
	lda NOTCH_TAB,x
	sta NOTCH_3

	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; bunches of tables to help compute modulo arithmetic, makes the coding 
;;; quite a bit simpler (and faster too) even if it does waste some space.
;;;  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

mod3:	
	.byte 0, 1, 2, 0, 1, 2

mod5:
	.byte 0, 1, 2, 3, 4
	.byte 0, 1, 2, 3, 4
mod6:
	.byte 0, 1, 2, 3, 4, 5
	.byte 0, 1, 2, 3, 4, 5
mod12:
	.byte 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11
	.byte 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11
mod26:
	.byte  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12
	.byte 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25
	.byte  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12
	.byte 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; jmp table designed to help drive the finite state machine.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

statelo:
	.byte <do_attract, <do_rotor, <do_ring 
	.byte <do_plug, <do_key, <do_encode, <do_perform_encode
	.byte <do_verify_rotor, <do_verify_plug, <do_error
statehi:
	.byte >do_attract, >do_rotor, >do_ring 
	.byte >do_plug, >do_key, >do_encode, >do_perform_encode
	.byte >do_verify_rotor, >do_verify_plug, >do_error

rowtab:	
	.byte 0, 0, 0, 0, 0, 0
	.byte 1, 1, 1, 1, 1, 1
	.byte 2, 2, 2, 2, 2, 2
	.byte 3, 3, 3, 3, 3, 3
	.byte 4, 4
coltab:
	.byte 0, 1, 2, 3, 4, 5
	.byte 0, 1, 2, 3, 4, 5
	.byte 0, 1, 2, 3, 4, 5
	.byte 0, 1, 2, 3, 4, 5
	.byte 2, 3

	.include "rotors.h"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; utility functions...
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	.advance $FFFC
	.word main
	.word main
